#!/usr/bin/perl -w

# $Id: ConfigLib.pm,v 1.15 2005/05/25 12:24:00 root Exp $

######################################################################
# NACE v0.2
#
# Copyright 2005 Russell Adams <rladams@adamsinfoserv.com>,
# and Adams Information Services.
#
# Licensed under GPL v2. See the README.
#
######################################################################

package ConfigLib;

######################################################################
# ConfigLib
#
# Contains functions used to manipulate hashes and snmp data.

use strict;

######################################################################
# Packaging options
require Exporter;

use vars qw( 
			 @ISA
			 @EXPORT
			 $Debug 
			 );

my @ISA = qw(Exporter);

my @EXPORT = qw( 
				 $Debug
				 &DebugMsg
				 &DebugDumpHash
				 &MergeHash
				 &AddHash
				 &CreateHashFromString
				 &PerformSubst
				 &GetConfigurationDirectives
				 &WriteRecord
				 );

# End Packaging Options
######################################################################


######################################################################
# DebugMsg

sub DebugMsg {

	$Debug && print STDERR "DEBUG: @_\n";

};

# End DebugMsg
######################################################################


######################################################################
# DebugDumpHash
#
# A debugging tool, dump the contents of a hash

sub DebugDumpHash {

	my ( $HashDesc , $HashPtr ) = @_ ;

	foreach ( sort keys %{$HashPtr} ) {

		defined $HashPtr->{$_} && 
			DebugMsg "$HashDesc:\t$_\t=>\t$HashPtr->{$_}";

	};

	DebugMsg "";

};

# End DebugDumpHash
######################################################################


######################################################################
# MergeHash
#
# Merge together two hashes, with overwrite.

sub MergeHash {
	
	# Returns a new HashC, created by taking HashA and HashB, and copying
	# each value to HashC, with HashB overwriting on copy.

	my ( $HashA , $HashB ) = @_ ;
	my %HashC;

	# Copy each key from HashA to HashC
	foreach ( keys %{$HashA} ) {

		$HashC{$_} = $HashA->{$_};

	};

	# Copy each key from HashB to HashC, overwriting existing values.
	foreach ( keys %{$HashB} ) {

		$HashC{$_} = $HashB->{$_};

	};

	return \%HashC;

};

# End MergeHash
######################################################################


######################################################################
# AddHash
#
# Merge together two hashes, any existing values in the second hash
# are appended to the value of the first as a comma separated list.

sub AddHash {

	# Returns a new HashC, created by copying HashA to HashC
	# and then appending any common entries from HashB to the
	# values in HashC.

	my ( $HashA , $HashB ) = @_ ;
	my %HashC;

	# Copy each key from HashA to HashC
	foreach ( keys %{$HashA} ) {

		$HashC{$_} = $HashA->{$_};

	};

	# Copy each key from HashB to HashC, appending existing values.
	foreach ( keys %{$HashB} ) {

		# Only if the value in HashB is defined.
		if ( defined $HashB->{$_} ) {
			
			if ( defined $HashC{$_} ) {
				
				$HashC{$_} .= "," . $HashB->{$_};
				
			} else {
				
				$HashC{$_} = $HashB->{$_};
				
			};
			
		};
		
	};
	
	return \%HashC;

};

# End AddHash
######################################################################


######################################################################
# CreateHashFromString
#
# Takes a string with colon delimited key=value pairs and returns
# a hash.

sub CreateHashFromString {

	my ( $ValidKeys , $RawString ) = @_ ;

	my %NewHash;

	foreach ( split( /:/ , $RawString ) ) {

		DebugMsg "CreateHashFromString Raw Key/Value: $_";

		# split out key/value pair
		my ( $Key , $Value ) = split(/=/);

		# If value field is borked, skip it.
		# This also cuts out the '^Nagios:' header 
		( ! defined $Key   ) && next;
		( ! defined $Value ) && next;

		# Purge leading or trailing whitespace
		$Key   =~ s/^\s+|\s+$//g;
		$Value =~ s/^\s+|\s+$//g;

		# Check keys, only add good ones.
		if ( $Key =~ m/$ValidKeys/ ) {
			$NewHash{ $Key } = "$Value";
			DebugMsg "CreateHashFromString Added valid pair: $Key = $Value";
		};
		
	};

	return \%NewHash;

};

# End CreateHashFromString
######################################################################


######################################################################
# PerformSubst
#
# Replaces arg2 with arg3 inside any value in given hash.
# Returns a NEW HASH.

sub PerformSubst {

	my ( $Hash , $Pattern , $Replacement ) = @_ ;

	my %NewHash;

	# Step through each key
	foreach ( keys %{$Hash} ) {

		# If its got a value, subst it via regexp
		if ( defined $Hash->{$_} ) {

			$NewHash{$_} = $Hash->{$_};
			$NewHash{$_} =~ s/$Pattern/$Replacement/;

		} else {

			$NewHash{$_} = undef;

		};

	};

	return \%NewHash;

};

# End PerformSubst
######################################################################


######################################################################
# GetConfigurationDirectives
#
# Use existing session to grab key/value pairs from sysContact
# and sysLocation fields.

sub GetConfigurationDirectives {

	# Import existing session
	my ( $SNMPSession , $MatchingList ) = @_ ;

	my ( $SNMPResponse, $HostData , $TempData);

	my %Oids = (
				'sysContact'     => '1.3.6.1.2.1.1.4.0',
				'sysLocation'    => '1.3.6.1.2.1.1.6.0',
				);
	
	# Grab certain oids to scrape data from
	foreach ( ( $Oids{'sysContact'} , $Oids{'sysLocation'} ) ) {

		########################################
		# Query oid
		DebugMsg "Fetching $_";

		$SNMPResponse = $SNMPSession->get_request( $_ );

		if ( ! defined $SNMPResponse ) {
			print STDERR $SNMPSession->error() . "\n";
			return undef;
		};
		
		if ( ! defined $SNMPResponse->{ $_ } ) {
			print STDERR $SNMPSession->error() . "\n";
			next;
		};
				
		DebugMsg "Retrived $_: " . $SNMPResponse->{ $_ };
		
		########################################
		# If the line begins with our tag then
		# Parse it and add it to our host data
		if ( $SNMPResponse->{ $_ } =~ m/^Nagios:/ ) {
			$TempData = CreateHashFromString( $MatchingList , 
											  $SNMPResponse->{$_}
											  );
			$HostData = AddHash( $HostData , $TempData );
		};

	};

	return $HostData;

};

# End GetConfigurationDirectives
######################################################################


######################################################################
# WriteRecord

sub WriteRecord {

	my ( $HostData , $RecordType , $MappingHash , $DestFile ) = @_ ;

	DebugDumpHash( 'Host' , $HostData );
	DebugDumpHash( 'Map' , $MappingHash );

	DebugMsg "Writing record to $DestFile";

	########################################
	# Write configuration for host

	open(MYFILE,"$DestFile")
		|| die "$!\n";

	DebugMsg "Writing host data to $DestFile";

	# Create record
	printf( MYFILE "define $RecordType {\n" );

	# Write out each key using mapping
	foreach ( sort keys %{$MappingHash} ) {

		if ( defined $HostData->{$_} ) {

			if ( $MappingHash->{$_} =~ /#/ ) {

				# Don't tab delimit a commented field
				printf( MYFILE "%s\t%s\n",
						$MappingHash->{$_},
						$HostData->{$_}
						);

			} else {

				# Tab the rest
				printf( MYFILE "\t%s\t%s\n",
						$MappingHash->{$_},
						$HostData->{$_}
						);

			};

		};

	};

	printf( MYFILE "}\n" );

	close MYFILE;

};

# End WriteRecord
######################################################################
