#!/usr/bin/perl -w

# $Id: GenerateHostGroups.pl,v 1.10 2005/05/08 02:09:36 root Exp $	

######################################################################
# NACE v0.2
#
# Copyright 2005 Russell Adams <rladams@adamsinfoserv.com>.
#
# Licensed under GPL v2.
#
# Given input, pulls hostgroup assignments for hosts from host
# definitions where there is a commented unofficial hostgroup field.
#
# Example:
#
# The following host definition would assign host 192.168.1.126 to
# both the WindowsServers and HomeGroup host groups.
#
# define host {
#         address 192.168.1.126
#         alias   John's Server
# #       host_group      WindowsServers,HomeGroup
#         host_name       192.168.1.126
#         use     HostTemplate
# }
#
# This is to compensate for Nagios' requirement that all member hosts
# be defined in the host group definition in a static manner.
#
# Given a set of default field values, this script reads definitions
# like those above and generates the host group definition
# dynamically. 
#
######################################################################


######################################################################
# Includes

# Use our configuration library, if not in the current directory,
# it's path had best be in the PERLLIB environment variable.
use lib "../lib";
use ConfigLib;

# Other requirements
use strict;
use Getopt::Long qw( :config no_ignore_case );

# End Includes
######################################################################


######################################################################
# Global vars

use vars qw( 
			 @Inputs
			 %Options
			 %Defaults
			 %HostGroupKeyMapping
			 );

# End Global vars
######################################################################


######################################################################
# Default Options

########################################
# Options

%Options = ( DisplayHelp     => 0,   # Do we show help and exit?
			 UseStdIn        => 0,	 # Read STDIN only with this option on
			 OutputDir       => '.',
			 );

########################################
# Defaults HostGroup Hash Values

# Basic hash with values that can be overridden via arguments
%Defaults = ( HG => undef,        # HostGroup is required
			  DS => '@@@HOSTGROUP@@@', # Description for the hostgroup w/ subst
			  CG => undef,        # ContactGroups
			  ML => undef,        # No member list by default
			  );

########################################
# Output key mapping

# maps each hash key to a nagios keyword.
%HostGroupKeyMapping = ( HG => 'hostgroup_name',
						 DS => 'alias',
						 CG => 'contact_groups',
						 ML => 'members',
						 );

# End of default options
######################################################################


######################################################################
# Main Program
######################################################################

ProcessOptions();

DebugMsg('Beginning file scanning...');

my ( @Hosts , $HostList , $HostGroupData );

# Check if we should read host list from STDIN
if ( $Options{'UseStdIn'} ) {

	DebugMsg('Reading host data from STDIN.');

	# Take our host definitions on STDIN and look for our default HostGroup
	@Hosts = ScanFile( *STDIN , $Defaults{'HG'} );
	
} else {

	DebugMsg "Scanning inputs @Inputs";
		
	my $dir;

	# Blank out the hosts list to prevent looping.
	@Hosts = ();

	# Scan each input directory

	foreach $dir ( @Inputs ) {

		DebugMsg "Scanning $dir";

		opendir( MYDIR , $dir ) || die "Failed to open $dir: $!";

		my @files;

		# Looking only for *.cfg
		@files = grep { /\.cfg$/ } readdir( MYDIR );

		closedir MYDIR;

		# Scan each config file found
		foreach ( @files ) {

			DebugMsg "Scanning $dir/$_";

			open( MYFILE, "$dir/$_" ) || die "Failed to open $dir/$_: $!";

			# Scan the current file for hosts in the hostgroup.
			my @TempHosts = ScanFile( *MYFILE , $Defaults{'HG'} );

			close MYFILE;

			DebugMsg "New Hosts @TempHosts";

			# Add hosts found to total host list
			push( @Hosts , @TempHosts );

		};

	};

};

DebugMsg "Total number of hosts: $#Hosts";

if ( $#Hosts ge 0 ) {

	foreach ( @Hosts ) {

		# Convert the Hosts array into a comma delimited list.
		$HostList .= $_ . ',' ;

	};

	# Remove trailing comma
	$HostList =~ s/,$//g;

	DebugMsg "Generated $HostList";

    # Perform substitution on the result.
	$HostGroupData = PerformSubst( \%Defaults , '\@\@\@HOSTGROUP\@\@\@' ,
								   $Defaults{'HG'} );
	
    # Assign member list
	$HostGroupData->{'ML'} = $HostList;
	
    # Write our hostgroup record
	WriteRecord( $HostGroupData , 'hostgroup' , \%HostGroupKeyMapping , 
				 "> $Options{'OutputDir'}/$Defaults{'HG'}.cfg" );
	
};

######################################################################
# End Main Program
######################################################################


######################################################################
# Functions
######################################################################


######################################################################
# ProcessOptions

sub ProcessOptions {

	GetOptions( ''          => \$Options{'UseStdIn'},
				'h'         => \$Options{'DisplayHelp'},
				'I=s'       => \@Inputs,
				'O=s'       => \$Options{'OutputDir'},
				'd=s'       => \%Defaults,
				'D'         => \$Debug,
				);

	if ( $Options{'DisplayHelp'} ) {

		DisplayHelp();

	};

	########################################
	# Debug dump of options

	$Debug && do {

		DebugDumpHash( 'Options'   , \%Options   );
		DebugDumpHash( 'Defaults'  , \%Defaults  );

	};

    ########################################
    # Check for required options

	$Defaults{'HG'} || die "A HostGroup name is required. (-d HG)\n";
	$Defaults{'CG'} || die "A ContactGroup is required. (-d CG)\n";

	DebugMsg 'Options processing complete.';

};

# End ProcessOptions
######################################################################


######################################################################
# ScanFile

sub ScanFile {

	DebugMsg "Scanning file";

	my ( $HostGroup , @HostNames , $CurrentHN , $CurrentHG , $HeaderFound );

	( *HostDataFile , $HostGroup ) = @_ ; 
	
	while ( <HostDataFile> ) {

		# Read each line in file, and chomp each, then match based on
		# line contents looking for host records.
		chomp;

		# Match first line in a host definition
		/define host \{/ && do {

			DebugMsg "Found header: $_";

			$CurrentHN = undef;
			$CurrentHG = undef;
			$HeaderFound = 1 ;
			next;
		};

		# Match hostname line from a definition, only inside a definition
		/host_name/ && $HeaderFound && do {

			DebugMsg "Found HostName: $_";

			$CurrentHN = ( split(/	/) )[2];
			next;
		};

		# Match a hostgroup only inside a definition
		/host_group/ && $HeaderFound && do {
			
			DebugMsg "Found HostGroup: $_";

			$CurrentHG = ( split(/	/) )[2];
			next;
		};

		# Match a terminator and add, only at the end of a definition
		/\}/ && $HeaderFound && do {

			DebugMsg "Found terminator: $_";

			# Close current definition
			$HeaderFound = 0;
			
			# If the host group found matches our HG
			# then add the host to the list.
			# Works with lists of groups too.

			DebugMsg "Testing $CurrentHG against $HostGroup";
			
			if ( $CurrentHG =~ /$HostGroup/ ) {

				DebugMsg "Adding $CurrentHN to member list.";

				# Add a hostname to the members list.
				push( @HostNames, $CurrentHN );

			};
					  
		};

	};

	DebugMsg "Returning member list";

	return @HostNames;

};

# End ScanFile
######################################################################


######################################################################
# DisplayHelp

sub DisplayHelp {

	print <<EOF;

GenerateHostGroups.pl

Part of the NACE suite of tools.

Licensed under GPL v2.

Copyright 2005, Russell Adams <rladams\@adamsinfoserv.com>.

Given input, pulls hostgroup assignments for hosts from host
definitions where there is a commented unofficial hostgroup field.

Example:

The following host definition would indicate host 192.168.1.126
belongs in both the WindowsServers and HomeGroup host groups.

define host {
	address		192.168.1.126
	alias		John's Server
#	host_group	WindowsServers,HomeGroup
	host_name	192.168.1.126
	use		HostTemplate
}

This is to compensate for Nagios' requirement that all member hosts
be defined in the host group definition in a static manner.

Given a set of default field values, this script reads definitions
like those above and generates the host group definition
dynamically.

Syntax:

	-I <input directory>

		What directory (or directories if specified multiple times) to
		scan for Nagios configuration files. Each file is scanned for
		commented hostgroup fields to parse and use to generate a
		member list for the hostgroup definition.

	-	Read host definitions from STDIN and parse. Suitable
		for piping commands to this script.

	-O <output directory>

		Where to save the hostgroup definition.

	-d <key=value pair>

		Valid keys:

			HG - HostGroup name **required**
			DS - Description of the HostGroup, defaults to HG value
			CG - ContactGroups for the HostGroup **required**

		The DS field can contain spaces with proper quoting. In that
		event, the entire key=value pair must be quoted.

		The CG field can contain multiple ContactGroups in a comma
		delimited list without spaces.

	-D	Debug, enables very verbose debugging information.

Sample Usage:

	GenerateHostGroups.pl \\
		-I ~nagios/etc/Auto/New \\
		-O ~nagios/etc/Auto/New \\
		-d HG=WindowsServers    \\
		-d 'DS=Windows Servers' \\
		-d CG=HomeAdmins

	The above syntax scans ~nagios/etc/Auto/New/*.cfg for host definitions
	with an commented hostgroup field that includes the WindowsServers
	host group. Note the quoting on the DS key=value pair.

	Each host found is added to the member list, the the host group
	definition is written to ~nagios/etc/Auto/New/WindowsServers.cfg,
	using the field values specified with -d, and the generated member
	list.

EOF

	exit -1;

};


# End DisplayHelp
######################################################################
