#!/usr/bin/perl -w

# $Id: QueryHosts.pl,v 1.15 2005/05/14 13:30:06 root Exp $	

######################################################################
# NACE v0.2
#
# Copyright 2005 Russell Adams <rladams@adamsinfoserv.com>,
# and Adams Information Services.
#
# Licensed under GPL v2. See the README.
#
######################################################################


######################################################################
# Includes

# Use our configuration library, if not in the current directory,
# it's path had best be in the PERLLIB environment variable.
use lib "../lib";
use ConfigLib;

# Other requirements
use strict;
use Net::SNMP; # Tested on v3.6, on Debian Woody.
use Getopt::Long qw( :config no_ignore_case );
use Parallel::ForkManager;

# End Includes
######################################################################


######################################################################
# Global vars

use vars qw( 
			 %Options
			 %Defaults
			 %Additions
			 %Oids
			 %HostKeyMapping
			 $ForkManager
			 );

%Oids = ( 'sysDescr'       => '1.3.6.1.2.1.1.1.0',
		  'sysContact'     => '1.3.6.1.2.1.1.4.0',
		  'sysLocation'    => '1.3.6.1.2.1.1.6.0',
		 );


# End Global vars
######################################################################


######################################################################
# Default Options

########################################
# Options

%Options = ( DisplayHelp     => 0,   # Do we show help and exit?
			 UseStdIn        => 0,	 # Read STDIN only with this option on
			 Community       => 'public', # Default community
			 Port            => '161',    # Default port
			 OutputDir       => '.', # Create files in current directory 
			 IncludeAll      => 0,	 # Don't include hosts w/o SNMP response 
			 IncludeMatching => 0,   # No default sysDescr regexp for inclusion
			 MaxForks        => 0,   # Don't do parallel by default yet
			 NoSNMPScan      => 0,   # Scan by default
			 );

########################################
# Defaults Host Hash Values

# Basic hash with values that can be optionally overridden via
# configuration options on the host.

%Defaults = ( HT => undef,        # HostTemplate to "use", required
			  HG => undef,        # HostGroup is required
			  DS => '@@@HOST@@@', # Description for the host w/ subst,
								  #   defaults to hostname
			  PN => undef,        # No default ParentNode
			  );

########################################
# Additional Options

%Additions = ( HG  => undef, # Additional HostGroups
			   PN  => undef, # Additional parents
			   );

########################################
# Output key mapping
# maps each host key to a nagios keyword.
%HostKeyMapping = ( HT => 'use',
					HN => 'host_name',
					AD => 'address',
					HG => '#	host_group', # Not a real nagiosfield
					DS => 'alias',
					PN => 'parents',
					);

# End of default options
######################################################################


######################################################################
# Main Program
######################################################################

ProcessOptions();

DebugMsg('Beginning SNMP scanning...');

# See if we're going parallel.
if ( $Options{'MaxForks'} > 0 ) {

	DebugMsg("Using Parallel Fork Manager, maximum $Options{'MaxForks'} children.");

	$ForkManager = new Parallel::ForkManager( $Options{'MaxForks'} );

};

# Check if we should read host list from STDIN
if ( $Options{'UseStdIn'} ) {

	DebugMsg('Reading hosts from STDIN.');

	# Read hosts from stdin
	while ( <STDIN> ) {

		# If we're parallel, engage a new fork.
		defined $ForkManager && do {
			my $pid = $ForkManager->start and next;
		};

		SNMPScanHost( chomp($_) );

		# End fork
		defined $ForkManager && $ForkManager->finish;

	};
	
} else {

	# Read hosts from args
	foreach ( @ARGV ) {
		
		# if we're parallel, create a new fork
		defined $ForkManager && do {
			my $pid = $ForkManager->start and next;
		};

        SNMPScanHost( $_ );

		# End fork
		defined $ForkManager && $ForkManager->finish;

    };  

};

# Wait til complete
defined $ForkManager && $ForkManager->wait_all_children;

######################################################################
# End Main Program
######################################################################


######################################################################
# Functions
######################################################################


######################################################################
# ProcessOptions

sub ProcessOptions {

	GetOptions( ''          => \$Options{'UseStdIn'},
				'I'         => \$Options{'UseStdIn'},
				'h'         => \$Options{'DisplayHelp'},
				'c=s'       => \$Options{'Community'},
				'p=i'       => \$Options{'Port'},
				'O=s'       => \$Options{'OutputDir'},
				'A'         => \$Options{'IncludeAll'},
				'M=s'       => \$Options{'IncludeMatching'},
				'd=s'       => \%Defaults,
				'a=s'       => \%Additions,
				'f=i'       => \$Options{'MaxForks'},
				'n'         => \$Options{'NoSNMPScan'},
				'D'         => \$Debug,
				);

	if ( $Options{'DisplayHelp'} ) {

		DisplayHelp();

	};	

	########################################
	# Debug dump of options

	$Debug && do {

		DebugDumpHash( 'Options'   , \%Options   );
		DebugDumpHash( 'Defaults'  , \%Defaults  );
		DebugDumpHash( 'Additions' , \%Additions );

	};

    ########################################
    # Check for required options

	$Defaults{'HG'} || die "A HostGroup is required. (-d HG)\n";
	$Defaults{'HT'} || die "A HostTemplate is required. (-d HT)\n";

    ########################################
    # Check for exclusive options

	( $Options{'IncludeAll'} && $Options{'IncludeMatching'} ) 
		&& die '-A and -M cannot be used together.\n';

	DebugMsg 'Options processing complete.';

};

# End ProcessOptions
######################################################################


######################################################################
# SNMPScanHost

sub SNMPScanHost {

	DebugMsg "Scanning $_";

	my ( $SNMPSession , $SNMPError, $SNMPResponse, 
		 $RemoteDescr, $RemoteData , $HostData );

	if ( ! $Options{'NoSNMPScan'} ) {

		# Create session object (does NOT open a connection, its stateless)
		# What about version, timeout, retries, debug mode?
		( $SNMPSession , $SNMPError ) = Net::SNMP->session( 
									   -hostname    => $_,
									   -community   => $Options{'Community'},
									   -port        => $Options{'Port'},
									   );

		# Check session object was created
		if ( ! defined $SNMPSession ) {
			print STDERR "$SNMPError\n";
			return undef;
		};
		
		# Query sysDescr
		$SNMPResponse = $SNMPSession->get_request( $Oids{'sysDescr'} );
		
	}

	# If not defined, there was no response.
	if ( ! defined $SNMPResponse ) {
		
		# Display verbose error information.
		$Options{'NoSNMPScan'} || print STDERR $SNMPSession->error() . "\n";

		# SNMP agent didn't respond, but if we IncludeAll, create
		# a default record for the host
		if ( $Options{'IncludeAll'} || $Options{'NoSNMPScan'} ) {

			DebugMsg "Due to IncludeAll or NoSNMPScan mode, creating entry for $_ anyway.";

			########################################
			# Create the host record
			
			# Add the Additions
			$HostData = AddHash( \%Defaults , \%Additions  );

			# Perform substitution on the result.
			$HostData = PerformSubst( $HostData , '\@\@\@HOST\@\@\@' , $_ );

			# Include the hostname
			$HostData->{'HN'} = "$_";
			$HostData->{'AD'} = "$_";

			# Write our host record
			WriteRecord( $HostData , 'host' , \%HostKeyMapping , 
						 "> $Options{'OutputDir'}/$_.cfg" );
			
		};

		# Leave function immediately.
		return undef;
			
	};

	############################################
	# SNMP responded, continue to evaluate host.
	$RemoteDescr = $SNMPResponse->{ $Oids{'sysDescr'} };
	
	DebugMsg "Retrived sysDescr of $_: $RemoteDescr";
	
	################################################
	# Test include regexp vs descr if in match mode
	# If we don't match the include regexp, skip
	if ( $Options{'IncludeMatching'} ) {

		if ( $RemoteDescr !~ /$Options{'IncludeMatching'}/ ) {
			
			# Host sysDescr didn't match inclusion regexp, skip

			DebugMsg "$_ did not match /$Options{'IncludeMatching'}/, skipping.";
			return undef;

			};

		DebugMsg "$_ matched /$Options{'IncludeMatching'}/, including.";

	};


	########################################
	# Query given remote tags
	$RemoteData = GetConfigurationDirectives( $SNMPSession , 'DS|PN|HG' );

	DebugDumpHash( "RemoteData for $_" , $RemoteData );

	# Done pulling configuration directives from remote host
	########################################################

	# Finished with SNMP.
	$SNMPSession->close();

	########################################
	# Create the host record

	# Override Defaults with RemoteData collected
	$HostData = MergeHash( \%Defaults , $RemoteData );

	# Add the Additions
	$HostData = AddHash(    $HostData , \%Additions  );

	# Perform substitution on the result.
	$HostData = PerformSubst( $HostData , '\@\@\@HOST\@\@\@' , $_ );

	# Include the hostname
	$HostData->{'HN'} = "$_";
	$HostData->{'AD'} = "$_";

	# Write our host record
	WriteRecord( $HostData , 'host' , \%HostKeyMapping , 
				 "> $Options{'OutputDir'}/$_.cfg" );

};

# End SNMPScanHost
######################################################################


######################################################################
# DisplayHelp

sub DisplayHelp {

	print <<EOF;

QueryHosts.pl

Part of the NACE suite of tools.

Licensed under GPL v2.

Copyright 2005, Russell Adams <rladams\@adamsinfoserv.com>.

Dynamically creates host definitions for Nagios using host
configuration field data given on the command line, and additionally
configuration directives fetched from the host.

Given a list of hosts via as arguments or via STDIN, uses SNMP to poll
each host and apply matching criteria to sysDescr to control whether
or not to create a host definition for the host. (ie: Only create host
definitions for Linux or Windows servers)

Additionally, uses SNMP to poll the contents of two custom fields of
the SNMP agent (sysContact and sysLocation) and examines them for
configuration directives in the form of a header and a sequence of
key=value pairs. These directives are then used when creating the host
definition to override default field values such as the HostGroup the
host belongs to, Description for the host, Parents for the host, etc.

In order to import configuration data from the SNMP Agent custom
fields on the remote host, custom fields must begin with 'Nagios:' and
match /^Nagios: /, then contain key=value pairs in a colon delimited
list. Whitespace is ignored.

The same keys used to specify the default values for fields in the
host definition on the command line are used in the custom
fields. These will override the default value for that field.

The idea behind storing configuration directives in the host SNMP
Agent allows the administrator of the remote host to specify how the
remote host should be setup in Nagios. This relieves the Nagios
administrator of the burden of manually coding the configuration for
the remote host. Furthermore, when applied across multiple hosts this
allows for automated configuration of groups of hosts. 

The Nagios administrator would then spend time writing "queries" to
setup hosts in groups instead of tediously coding configurations for
every host to be monitored. Some example queries could work via
grepping a domain transfer looking for hosts matching a pattern, or
nmap scanning a subnet to find hosts to monitor.

Example fields from SNMP Agents with Nagios configuration directives:

system.sysLocation.0 = Nagios: DS=John\'s Server : HG=WindowsServers

In the above example, the host description would become 
"John's Server", and the host would include a field to include it 
in the WindowsServers HostGroup.

Syntax:

	-I or -

		Read hostnames to poll from STDIN. Otherwise hostnames are
		given as arguments to the script.

	-O <output directory>

		Where to save the host definitions (ie: Hostname.cfg) 

		The output directory is intended to be included in the Nagios
		configuration file (nagios.cfg) using the cfgdir option in
		order to process all of the host definitions created in
		individual files.

	-c <community>

		SNMP community to use when polling, defaults to public.

	-p <port>

		SNMP port to use when polling, defaults to 169.

    -f <forks>

        Maximum number of hosts to scan simultaneously. Basically max
        number of child processes or forks.

	-A

		Create host definitions for ALL hosts whether they were
		successfully polled or not. Hosts not responding to SNMP use
		the defaults for all fields.

	-M <regexp>

		Create host definitions for hosts whose sysDescr MATCH the
		specified regular expression.

	-D	Debug, enables very verbose debugging information.

	-d <key=value pair>

		Default values for fields in the host definition.

		Valid keys:

			HT - HostTemplate to 'use' **required**
			HG - HostGroup **required**
			DS - Description for the host, defaults to the host name
			PN - Parent node(s) for the host, comma separated.

		The HT field refers to the HostTemplate the host definition
		should inherit from with the Nagios 'use' directive. This
		script relies heavily upon the template method of
		configuration.

		The HG field does not correspond to any official Nagios host
		fields, but creates a commented pseudo-field called host_group
		which is later parsed by the GenerateHostGroups.pl script
		which creates host group definitions dynamically.

		The DS field is the host alias, and can contain spaces with
		proper quoting. In that event, the entire key=value pair must
		be quoted.

		The PN field can contain multiple parents in a comma delimited
		list without spaces.

	-a <key=value pair>

		Additional field values appended to any polled remote values
		or the defaults upon the creation of the host definition.

		Valid keys:

			HG - Additional HostGroups
			PN - Additional Parent node(s) for the host

		The additional values are useful to include an extra value
		outside of the defaults or polled remote data. For instance,
		the default HostGroup may be WindowsServers, the polled remote
		value of WebServers would override the default. However, if
		WindowsServers is specified as an additional HostGroup, then
		the final list HostGroups would be
		"Webservers,WindowsServers". Duplicate HostGroups have no
		negative effect in this regard were the default value used.

Sample Usage:

	nmap -sP 192.168.1.* -oG -              \\
		| awk '/Status: Up/ { print $2 }'   \\
		| QueryHosts.pl                     \\
		-O ~nagios/Auto/New                 \\
		-d HT=HostTemplate                  \\
		-d HG=HomeGroup                     \\
		-a HG=HomeGroup                     \\
		-A                                  \\
		-

	The above example uses nmap to do a ping scan of the 192.168.1.0
	subnet and output it in machine parsable mode. Awk pulls out any
	IP addresses that are up. The list of IP addresses is passed to
	the QueryHosts.pl script on STDIN.

	Host definitions are output to ~nagios/Auto/New/HostName.cfg, and
	told to 'use' HostTemplate. They include a comment indicating that
	they should belong to the HomeGroup HostGroup, and a definition is
	output for ALL hosts whether they responded to SNMP or not.

EOF

	exit -1;

};

# End DisplayHelp
######################################################################
