#!/usr/bin/perl -w

# $Id: QueryServices.pl,v 1.12 2005/05/08 02:05:43 root Exp $	

######################################################################
# NACE v0.2
#
# Copyright 2005 Russell Adams <rladams@adamsinfoserv.com>.
#
# Licensed under GPL v2. See the README.
#
######################################################################

######################################################################
# Includes

# Use our configuration library, if not in the current directory,
# it's path had best be in the PERLLIB environment variable.
use lib "../lib";
use ConfigLib;

# Other requirements
use strict;
use Net::SNMP; # Tested on v3.6, on Debian Woody.
use Getopt::Long qw( :config no_ignore_case );
use Digest::MD5 qw( md5_hex );
use Parallel::ForkManager;

# End Includes
######################################################################


######################################################################
# Global vars

use vars qw( 
			 %Options
			 %Defaults
			 %Additions
			 %Oids
			 %Services
			 @Targets
			 %ServiceKeyMapping
			 $ForkManager
			 );

%Oids = ( 'sysDescr'       => '1.3.6.1.2.1.1.1.0',
		  'sysContact'     => '1.3.6.1.2.1.1.4.0',
		  'sysLocation'    => '1.3.6.1.2.1.1.6.0',
		  );


# End Global vars
######################################################################


######################################################################
# Default Options

########################################
# Options

%Options = ( DisplayHelp     => 0,   # Do we show help and exit?
			 UseStdIn        => 0,	 # Read STDIN only with this option on
			 Community       => 'public', # Default community
			 Port            => '161',    # Default port
			 OutputDir       => '.', # Create files in current directory 
			 MaxForks        => 0,   # Don't do parallel by default yet
			 );

########################################
# Defaults Host Hash Values

# Basic hash with values that can be optionally overridden via
# configuration options on the host.

%Defaults = ( ST => undef,          # ServiceTemplate to "use", required
			  CG => undef,          # ContactGroup is required
			  SD => '@@@TARGET@@@', # Description for the service
			  SC => undef,          # Service check command, required
			  );

########################################
# Additional Options

%Additions = ( CG  => undef, # Additional ContactGroups
			   );

########################################
# Service Options

%Services = ( Key     => undef,         # No key by default
			  RegExp  => undef,         # No regexp by default
			  OID     => undef,         # No oid by default
			  Exec    => undef,         # No program to run by default
			  );

########################################
# Output key mapping
# maps each host key to a nagios keyword.
%ServiceKeyMapping = ( ST => 'use',
					   SD => 'service_description',
					   SC => 'check_command',
					   CG => 'contact_groups',
					   HN => 'host_name',
					   );

# End of default options
######################################################################


######################################################################
# Main Program
######################################################################

ProcessOptions();

DebugMsg('Beginning SNMP scanning...');

# See if we're going parallel.
if ( $Options{'MaxForks'} > 0 ) {

	$ForkManager = new Parallel::ForkManager( $Options{'MaxForks'} );

};

# Check if we should read host list from STDIN
if ( $Options{'UseStdIn'} ) {

	DebugMsg('Reading hosts from STDIN.');

	# Read hosts from stdin
	while ( <STDIN> ) {

		# If we're parallel, engage a new fork.
		defined $ForkManager && do {
			my $pid = $ForkManager->start and next;
		};

		chomp;
		
		ServiceScanHost( $_ );

		# End fork
		defined $ForkManager && $ForkManager->finish;
		
	};
	
} else {

	# Read hosts from args
	foreach ( @ARGV ) {
		
		# if we're parallel, create a new fork
		defined $ForkManager && do {
			my $pid = $ForkManager->start and next;
		};

        ServiceScanHost( $_ );

		# End fork
		defined $ForkManager && $ForkManager->finish;

    };  

};

# Wait til complete
defined $ForkManager && $ForkManager->wait_all_children;

######################################################################
# End Main Program
######################################################################


######################################################################
# Functions
######################################################################


######################################################################
# ProcessOptions

sub ProcessOptions {

	GetOptions( ''          => \$Options{'UseStdIn'},
				'I'         => \$Options{'UseStdIn'},
				'h'         => \$Options{'DisplayHelp'},
				'c=s'       => \$Options{'Community'},
				'p=i'       => \$Options{'Port'},
				'O=s'       => \$Options{'OutputDir'},
				'd=s'       => \%Defaults,
				'a=s'       => \%Additions,
				's=s'       => \%Services,
				'f=i'       => \$Options{'MaxForks'},
				'D'         => \$Debug,
				);

	if ( $Options{'DisplayHelp'} ) {

		DisplayHelp();

	};	

	########################################
	# Debug dump of options

	$Debug && do {

		DebugDumpHash( 'Options'   , \%Options   );
		DebugDumpHash( 'Defaults'  , \%Defaults  );
		DebugDumpHash( 'Additions' , \%Additions );
		DebugDumpHash( 'Services' , \%Additions );

	};

    ########################################
    # Check for required options

	$Defaults{'CG'} || die "A ContactGroup is required. (-d CG)\n";
	$Defaults{'ST'} || die "A ServiceTemplate is required. (-d ST)\n";
	$Defaults{'SC'} || die "A ServiceCheckCommand is required. (-d SC)\n";

    ########################################
    # Check for exclusive options

	( defined $Services{'Key'} && 
	  ( defined $Services{'RegExp'} || defined $Services{'OID'} ) )
		&& die 'A Service Key and RegExp/OID pair are exclusive.\n';
									
	( defined $Services{'Exec'} && 
	  ( defined $Services{'RegExp'} || defined $Services{'OID'} ) )
		&& die 'Exec and RegExp/OID pair are exclusive.\n';

	( defined $Services{'Key'} && 
	  defined $Services{'Exec'} )
		&& die 'A Service Key and Exec are exclusive.\n';

	( defined $Services{'Key'}
	  || defined $Services{'RegExp'}
	  || defined $Services{'Exec'} ) 
		|| die "Please specify how to create services.\n";
		
	DebugMsg 'Options processing complete.';

};

# End ProcessOptions
######################################################################


######################################################################
# ServiceScanHost

sub ServiceScanHost {

	DebugMsg "Scanning $_";

	my ( $HostName ) = @_ ;

	# So are we fetching the value of a key, walking and matching an
	# an oid, or execing a program to get our target list?


	# We always contact the host to get configuration directives
	# However, if they are unavailable it is not a problem.

	my ( $SNMPSession , $SNMPError, $SNMPResponse, 
		 $RemoteDescr, $RemoteData , $ServiceData , $FileName );
	
	$#Targets = -1;

	# Create session object (does NOT open a connection, its stateless)
    # What about version, timeout, retries, debug mode?
	( $SNMPSession , $SNMPError ) = Net::SNMP->session( 
									   -hostname    => $HostName,
									   -community   => $Options{'Community'},
									   -port        => $Options{'Port'},
									   );

	# Check session object was created
	if ( ! defined $SNMPSession ) {
		print STDERR "$SNMPError\n";
		return undef;
	};

	# Query sysDescr
	$SNMPResponse = $SNMPSession->get_request( $Oids{'sysDescr'} );
	
	# If not defined, there was no response.
	if ( defined $SNMPResponse ) {
		
		DebugMsg "Received sysDescr from $HostName: $SNMPResponse->{ $Oids{'sysDescr'} }";
		
		########################################
		# Query given remote tags
		
		if ( $Services{'Key'} ) {

			DebugMsg "Fetching configuration directives with key.";

			# If we've been given a service key to fetch, get it
			$RemoteData = GetConfigurationDirectives( $SNMPSession ,
													  "CG|" . $Services{'Key'}
													  );

		} else {

			DebugMsg "Fetching basic configuration directives.";

			# Otherwise just grab our basic data
			$RemoteData = GetConfigurationDirectives( $SNMPSession ,
													  "CG"
													  );

		};

		DebugMsg "Done fetching directives.";
		
		DebugDumpHash( "RemoteData for $HostName" , $RemoteData );
		
		# Done pulling configuration directives from remote host
		########################################################
		
	};

	########################################
	# Now handle each service option

	# Services by Key
	if ( $Services{'Key'} ) {

		# If we're supposed to grab out target list from a remote
		# directive, it was already fetched. Just copy it in place.
		
		if ( defined $RemoteData->{ $Services{'Key'} } ) {

			DebugDumpHash( 'RemoteData' , $RemoteData );

			@Targets = split(/,/ , $RemoteData->{ $Services{'Key'} } );
			
			$FileName = "$HostName" . "_$Services{'Key'}.cfg"

			} else {
				
				# No key on remote end
				
				return undef;

			};
			
	};

	# Services by Exec, not entirely safe...
	if ( $Services{'Exec'} ) {

		# Take each line of output as a target
		my $Command = $Services{'Exec'};
		$Command =~ s/\@\@\@HOST\@\@\@/$HostName/g;

		chomp( @Targets = `$Command` );

		$FileName = "$HostName" . "_" . md5_hex( $Services{'Exec'} ) . ".cfg";

	};

	if ( $Services{'RegExp'} ) {

		$FileName = "$HostName" . "_$Services{'OID'}.cfg";

		# Pull snmp table of OID, and set targets as matching entries.
		$SNMPResponse = 
			$SNMPSession->get_table( $Services{'OID'} ) 
			|| do {
				print STDERR $SNMPSession->error() . "\n";
				return undef;
			};
	
		DebugMsg "Retrived SNMP table $Services{'OID'} to scan." ;
		DebugDumpHash ( 'Table' , $SNMPResponse );
		
		foreach ( keys %$SNMPResponse ) {

			DebugMsg "Testing $SNMPResponse->{$_} against /$Services{'RegExp'}/";

			$SNMPResponse->{ $_ } =~ /$Services{'RegExp'}/ || next;
			
			DebugMsg "Adding target for $SNMPResponse->{$_}.";
			
			my $SanitizedName = $SNMPResponse->{ $_ };
			$SanitizedName =~ tr/\\!\#$`/./;
			
			push( @Targets , $SanitizedName );

		};

	};

	# Finished with SNMP.
	$SNMPSession->close();

	########################################
	# Create the service records

	foreach ( @Targets ) {

		# Override Defaults with RemoteData collected
		$ServiceData = MergeHash( \%Defaults , $RemoteData );
		
		# Add the Additions
		$ServiceData = AddHash( $ServiceData , \%Additions  );
		
		# Perform substitution on the result.
		$ServiceData = PerformSubst( $ServiceData ,
									 '\@\@\@HOST\@\@\@' ,
									 $HostName );

		$ServiceData = PerformSubst( $ServiceData ,
									 '\@\@\@TARGET\@\@\@' ,
									 $_ );

		# Include the hostname
		$ServiceData->{'HN'} = "$HostName";
		
		# Write our host record
		WriteRecord( $ServiceData , 'service' , \%ServiceKeyMapping , 
					 ">> $Options{'OutputDir'}/" . $FileName );
		
	};

};

# End SNMPScanHost
######################################################################


######################################################################
# DisplayHelp

sub DisplayHelp {

	print <<EOF;

QueryServices.pl

Part of the NACE suite of tools.

Licensed under GPL v2.

Copyright 2005, Russell Adams <rladams\@adamsinfoserv.com>.

Dynamically creates service definitions for Nagios using server
configuration field data given on the command line, and additionally
configuration directives fetched from the target host.

QueryServices.pl builds on the automation ideas of QueryHosts.pl,
enabling the Nagios administrator to write queries to monitor
interfaces on routers, all the disks on a host, specific processes,
etc.

This script builds a list of targets to create services for. Targets
could be disk names, interface numbers, process names, etc. Once a
list of targets has been gathered, a service definition is created for
each target where the target is substituted into the check command
arguments.

Given a list of hosts via as arguments or via STDIN, uses SNMP to poll
each host for additional configuration fields. Then, uses one of three
methods to identify targets for the host.

The first method is via a service key, which would be downloaded with
the configuration data from the remote host via SNMP. For example,
using a service key of 'PL' (ie: -s Key=PL), the value of the 'PL' key
from the remote host would be a comma delimited list of targets to
create service definitions with. In this example, each item in the
value of the remote 'PL' key would be a process name, and sent to the
check command as ARG1 in each service definition.

The second method to generate a list of targets is to run an arbitrary
command and read each line returned as a target (ie: -s 'Exec=grep
\@\@\@HOST\@\@\@ filename'). Substitution is performed on the command in
advance, replacing '\@\@\@HOST\@\@\@' with the host name. This could be used
for database lookups, a grep of a file, or even a detailed snmpwalk of
a host looking for specific criteria.

The third and last method for generating a list of targets is to
specify an SNMP OID in numeric form (ie: -s
OID=1.3.6.1.2.1.25.2.3.1.3) and a regexp used to match against the
results (ie: -s 'RegExp=.*' (match all) ). The OID is walked with the
Net::SNMP get_table function, and the results that match the RegExp
are designated as targets.

Syntax:

	-I or -

		Read hostnames to poll from STDIN. Otherwise hostnames are
		given as arguments to the script.

	-O <output directory>

		Where to save the host definitions (ie: Hostname.cfg) 

		The output directory is intended to be included in the Nagios
		configuration file (nagios.cfg) using the cfgdir option in
		order to process all of the host definitions created in
		individual files.

	-c <community>

		SNMP community to use when polling, defaults to public.

	-p <port>

		SNMP port to use when polling, defaults to 169.

	-f <forks>

        Maximum number of hosts to scan simultaneously. Basically max
        number of child processes or forks.

	-D	Debug, enables very verbose debugging information.

	-d <key=value pair>

		Default values for fields in the host definition.

		Valid keys:

			ST - ServiceTemplate to 'use' **required**
			CG - ContactGroup(s) **required**
			SD - Description for the service, defaults to the target name
			SC - Service check command **required**

		The ST field refers to the ServiceTemplate the host definition
		should inherit from with the Nagios 'use' directive.

		The CG field is the ContactGroup(s) for the service.

		The SD field is the service description, and can contain
		spaces with proper quoting. In that event, the entire
		key=value pair must be quoted. '\@\@\@TARGET\@\@\@' is replaced in
		the value with the target namw.

		The SC field is the service check command, and can contain
		spaces with proper quoting. Like SD, '\@\@\@TARGET\@\@\@' is
		replaced with the target name.

	-a <key=value pair>

		Additional field values appended to any polled remote values
		or the defaults upon the creation of the host definition.

		Valid keys:

			CG - Additional ContactGroups

	-s <key=value pair>

		Method to build target list with and parameters.

		Valid keys:

			Key    - Use a remote value of Key on remote for target list.
			Exec   - Execute command and use output for target list.
			OID    - SNMP walk this OID, requires RegExp.
			RegExp - RegExp used to match targets in results from SNMP walk.

Sample Usage:

	awk '/host_name/ {print $2}' ~nagios/Auto/New/*.cfg            \\
	| sort                                                         \\
	| uniq                                                         \\
	| QueryServices.pl                                             \\
		-O ~nagios/Auto/New                                        \\
		-d ST=DiskServiceTemplate                                  \\
		-d CG=HomeAdmins                                           \\
		-d 'SC=CheckSNMPDisk!^\@\@\@TARGET\@\@\@$!80!90'                 \\
		-s OID=1.3.6.1.2.1.25.2.3.1.3                              \\
		-s 'RegExp=(^\/(?!dev)|^[A-Z]:.*Label.*Serial)'            \\
		-

	Given a directory of host definitions (optimally created with
	QueryHosts.pl), filter out all the host names, sort them and
	ensure they are unique. Feed the resulting list of hosts to
	QueryServices.pl.

	Output service definitions to
	~nagios/Auto/New/HostName_ServiceType.cfg. Use ServiceTemplate,
	contact HomeAdmins, and use 'CheckSNMPDisk...' for the check
	command with substitution.

	Search for targets by snmpwalking 1.3.6.1.2.1.25.2.3.1.3 (Host
	Resources MIB, hrStorageDescr - name of the remote disk) and match
	any values returned against /^\/(?!dev)|^[A-Z]:.*Label.*Serial/ to
	create our list of targets. This nasty regexp will match any UNIX
	pathnames except /dev*, and any Windows disk names with a serial
	number and label, which tend to be hard drives.

EOF

	exit -1;

};

# End DisplayHelp
######################################################################
